"""
This module defines the SArray class which provides the
ability to create, access and manipulate a remote scalable array object.

SArray acts similarly to pandas.Series but without indexing.
The data is immutable, homogeneous, and is stored on the GraphLab Server side.
"""
import graphlab.connect as _mt
import graphlab.connect.main as glconnect
from graphlab.cython.cy_type_utils import pytype_from_dtype, infer_type_of_list, is_numeric_type
from graphlab.cython.cy_sarray import UnitySArrayProxy
from graphlab.cython.context import debug_trace as cython_context
from graphlab.util import make_internal_url
import graphlab as gl
import graphlab.canvas
import graphlab.canvas.inspect
import graphlab.canvas.views.sarray
import inspect
import math
from graphlab.deps import numpy, HAS_NUMPY
from graphlab.deps import pandas, HAS_PANDAS
import time
import array
import datetime
import graphlab.meta as meta
import itertools
import warnings

__all__ = ['SArray']

def _create_sequential_sarray(size, start=0, reverse=False):
    if type(size) is not int:
        raise TypeError("size must be int")

    if type(start) is not int:
        raise TypeError("size must be int")

    if type(reverse) is not bool:
        raise TypeError("reverse must me bool")

    with cython_context():
        return SArray(_proxy=glconnect.get_unity().create_sequential_sarray(size, start, reverse))

class SArray(object):
    """
    An immutable, homogeneously typed array object backed by persistent storage.

    SArray is scaled to hold data that are much larger than the machine's main
    memory. It fully supports missing values and random access. The
    data backing an SArray is located on the same machine as the GraphLab
    Server process. Each column in an :py:class:`~graphlab.SFrame` is an
    SArray.

    Parameters
    ----------
    data : list | numpy.ndarray | pandas.Series | string
        The input data. If this is a list, numpy.ndarray, or pandas.Series,
        the data in the list is converted and stored in an SArray.
        Alternatively if this is a string, it is interpreted as a path (or
        url) to a text file. Each line of the text file is loaded as a
        separate row. If ``data`` is a directory where an SArray was previously
        saved, this is loaded as an SArray read directly out of that
        directory.

    dtype : {None, int, float, str, list, array.array, dict}, optional
        The data type of the SArray. If not specified (None), we attempt to
        infer it from the input. If it is a numpy array or a Pandas series, the
        dtype of the array/series is used. If it is a list, the dtype is
        inferred from the inner list. If it is a URL or path to a text file, we
        default the dtype to str.

    ignore_cast_failure : bool, optional
        If True, ignores casting failures but warns when elements cannot be
        casted into the specified dtype.

    Notes
    -----
    - If ``data`` is pandas.Series, the index will be ignored.
    - When working with the GraphLab EC2 instance (see
      :py:func:`graphlab.aws.launch_EC2()`), an SArray cannot be constructed
      using local file path, because it involves a potentially large amount of
      data transfer from client to server. However, it is still okay to use a
      remote file path. See the examples below. The same restriction applies to
      :py:class:`~graphlab.SGraph` and :py:class:`~graphlab.SFrame`.

    Examples
    --------
    >>> sa = SArray(data=[1,2,3,4,5], dtype=int)
    >>> sa = SArray('http://s3-us-west-2.amazonaws.com/testdatasets/a_to_z.txt.gz')
    >>> sa = SArray([[1,2,3], [3,4,5]])
    >>> sa = SArray(data=[{'a':1, 'b': 2}, {'b':2, 'c': 1}])

    Working with a GraphLab EC2 instance:

    >>> graphlab.aws.launch_EC2('m1.large')
    >>> sa = SArray('~/mydata/foo.csv')          # throws exception
    >>> sa = SArray('remote:///mydata/foo.csv')  # works
    >>> sa = SArray("http://testdatasets.s3-website-us-west-2.amazonaws.com/users.csv.gz") # works
    >>> sa = SArray("s3://mybucket/foo.csv") # works
    >>> graphlab.aws.terminate_EC2()
    """

    def __init__(self, data=[], dtype=None, ignore_cast_failure=False, _proxy=None):
        """
        __init__(data=list(), dtype=None, ignore_cast_failure=False)

        Construct a new SArray. The source of data includes: list,
        numpy.ndarray, pandas.Series, and urls.
        """
        _mt._get_metric_tracker().track('sarray.init')
        if dtype is not None and type(dtype) != type:
            raise TypeError('dtype must be a type, e.g. use int rather than \'int\'')

        if (_proxy):
            self.__proxy__ = _proxy
        elif type(data) == SArray:
            self.__proxy__ = data.__proxy__
        else:
            self.__proxy__ = UnitySArrayProxy(glconnect.get_client())
            # we need to perform type inference
            if dtype is None:
                if (isinstance(data, list)):
                    # if it is a list, Get the first type and make sure
                    # the remaining items are all of the same type
                    dtype = infer_type_of_list(data)
                elif isinstance(data, array.array):
                    dtype = infer_type_of_list(data)
                elif HAS_PANDAS and isinstance(data, pandas.Series):
                    # if it is a pandas series get the dtype of the series
                    dtype = pytype_from_dtype(data.dtype)
                    if dtype == object:
                        # we need to get a bit more fine grained than that
                        dtype = infer_type_of_list(data)

                elif HAS_NUMPY and isinstance(data, numpy.ndarray):
                    # if it is a numpy array, get the dtype of the array
                    dtype = pytype_from_dtype(data.dtype)
                    if dtype == object:
                        # we need to get a bit more fine grained than that
                        dtype = infer_type_of_list(data)
                    if len(data.shape) == 2:
                        # we need to make it an array or a list
                        if dtype == float or dtype == int:
                            dtype = array.array
                        else:
                            dtype = list
                    elif len(data.shape) > 2:
                        raise TypeError("Cannot convert Numpy arrays of greater than 2 dimensions")

                elif (isinstance(data, str)):
                    # if it is a file, we default to string
                    dtype = str

            if HAS_PANDAS and isinstance(data, pandas.Series):
                with cython_context():
                    self.__proxy__.load_from_iterable(data.values, dtype, ignore_cast_failure)
            elif (HAS_NUMPY and isinstance(data, numpy.ndarray)) or isinstance(data, list) or isinstance(data, array.array):
                with cython_context():
                    self.__proxy__.load_from_iterable(data, dtype, ignore_cast_failure)
            elif (isinstance(data, str)):
                internal_url = make_internal_url(data)
                with cython_context():
                    self.__proxy__.load_autodetect(internal_url, dtype)
            else:
                raise TypeError("Unexpected data source. " \
                                "Possible data source types are: list, " \
                                "numpy.ndarray, pandas.Series, and string(url)")

    @classmethod
    def from_const(cls, value, size):
        """
        Constructs an SArray of size with a const value.

        Parameters
        ----------
        value : [int | float | str | array.array | list | dict]
          The value to fill the SArray
        size : int
          The size of the SArray

        Examples
        --------
        Construct an SArray consisting of 10 zeroes:

        >>> graphlab.SArray.from_const(0, 10)
        """
        assert type(size) is int and size > 0, "size must be a positive int"
        if (type(value) not in set([int, float, str, array.array, list, dict])):
            raise TypeError('Cannot create sarray of value type %s' % str(type(value)))
        proxy = UnitySArrayProxy(glconnect.get_client())
        proxy.load_from_const(value, size)
        return cls(_proxy=proxy)

    @classmethod
    def from_avro(cls, url):
        """
        Construct an SArray from an Avro file. The SArray type is determined by
        the schema of the Avro file.

        Parameters
        ----------
        url : str
          The file to load into an SArray.

        References
        ----------
        - `Avro Specification <http://avro.apache.org/docs/1.7.7/spec.html>`_
        """
        proxy = UnitySArrayProxy(glconnect.get_client())
        proxy.load_from_avro(url)
        return cls(_proxy = proxy)

    def __get_content_identifier__(self):
        """
        Returns the unique identifier of the content that backs the SArray

        Notes
        -----
        Meant for internal use only.
        """
        with cython_context():
            return self.__proxy__.get_content_identifier()

    def save(self, targetfile):
        """
        Saves the SArray to file.

        The saved SArray will be in a directory named with the `targetfile` parameter.

        Parameters
        ----------
        targetfile: string
            The directory to save the SArray. Either a local path or a remote
            URL. If `targetfile` is an existing directory, `save` will fail if
            the directory is not a GraphLab archive (directory for previously
            saved GraphLab Create objects), else it will overwrite the existing
            directory.
        """
        with cython_context():
            self.__proxy__.save(make_internal_url(targetfile))

    def __repr__(self):
        """
        Returns a string description of the SArray.
        """
        ret =       "dtype: " + str(self.dtype().__name__) + "\n"
        ret = ret + "Rows: " + str(self.size()) + "\n"
        ret = ret + str(self)
        return ret

    def __str__(self):
        """
        Returns a string containing the first 100 elements of the array.
        """
        headln = str(list(self.head(100)))
        if (self.size() > 100):
            # cut the last close bracket
            # and replace it with ...
            headln = headln[0:-1] + ", ... ]"
        return headln

    def __nonzero__(self):
        """
        Returns true if the array is not empty.
        """
        return self.size() != 0

    def __len__(self):
        """
        Returns the length of the array
        """
        return self.size()

    def __iter__(self):
        """
        Provides an iterator to the contents of the array.
        """
        def generator():
            elems_at_a_time = 262144
            self.__proxy__.begin_iterator()
            ret = self.__proxy__.iterator_get_next(elems_at_a_time)
            while(True):
                for j in ret:
                    yield j

                if len(ret) == elems_at_a_time:
                    ret = self.__proxy__.iterator_get_next(elems_at_a_time)
                else:
                    break

        return generator()

    def __add__(self, other):
        """
        If other is a scalar value, adds it to the current array, returning
        the new result. If other is an SArray, performs an element-wise
        addition of the two arrays.
        """
        with cython_context():
            if type(other) is SArray:
                return SArray(_proxy = self.__proxy__.vector_operator(other.__proxy__, '+'))
            else:
                return SArray(_proxy = self.__proxy__.left_scalar_operator(other, '+'))

    def __sub__(self, other):
        """
        If other is a scalar value, subtracts it from the current array, returning
        the new result. If other is an SArray, performs an element-wise
        subtraction of the two arrays.
        """
        with cython_context():
            if type(other) is SArray:
                return SArray(_proxy = self.__proxy__.vector_operator(other.__proxy__, '-'))
            else:
                return SArray(_proxy = self.__proxy__.left_scalar_operator(other, '-'))

    def __mul__(self, other):
        """
        If other is a scalar value, multiplies it to the current array, returning
        the new result. If other is an SArray, performs an element-wise
        multiplication of the two arrays.
        """
        with cython_context():
            if type(other) is SArray:
                return SArray(_proxy = self.__proxy__.vector_operator(other.__proxy__, '*'))
            else:
                return SArray(_proxy = self.__proxy__.left_scalar_operator(other, '*'))

    def __div__(self, other):
        """
        If other is a scalar value, divides each element of the current array
        by the value, returning the result. If other is an SArray, performs
        an element-wise division of the two arrays.
        """
        with cython_context():
            if type(other) is SArray:
                return SArray(_proxy = self.__proxy__.vector_operator(other.__proxy__, '/'))
            else:
                return SArray(_proxy = self.__proxy__.left_scalar_operator(other, '/'))

    def __lt__(self, other):
        """
        If other is a scalar value, compares each element of the current array
        by the value, returning the result. If other is an SArray, performs
        an element-wise comparison of the two arrays.
        """
        with cython_context():
            if type(other) is SArray:
                return SArray(_proxy = self.__proxy__.vector_operator(other.__proxy__, '<'))
            else:
                return SArray(_proxy = self.__proxy__.left_scalar_operator(other, '<'))

    def __gt__(self, other):
        """
        If other is a scalar value, compares each element of the current array
        by the value, returning the result. If other is an SArray, performs
        an element-wise comparison of the two arrays.
        """
        with cython_context():
            if type(other) is SArray:
                return SArray(_proxy = self.__proxy__.vector_operator(other.__proxy__, '>'))
            else:
                return SArray(_proxy = self.__proxy__.left_scalar_operator(other, '>'))


    def __le__(self, other):
        """
        If other is a scalar value, compares each element of the current array
        by the value, returning the result. If other is an SArray, performs
        an element-wise comparison of the two arrays.
        """
        with cython_context():
            if type(other) is SArray:
                return SArray(_proxy = self.__proxy__.vector_operator(other.__proxy__, '<='))
            else:
                return SArray(_proxy = self.__proxy__.left_scalar_operator(other, '<='))


    def __ge__(self, other):
        """
        If other is a scalar value, compares each element of the current array
        by the value, returning the result. If other is an SArray, performs
        an element-wise comparison of the two arrays.
        """
        with cython_context():
            if type(other) is SArray:
                return SArray(_proxy = self.__proxy__.vector_operator(other.__proxy__, '>='))
            else:
                return SArray(_proxy = self.__proxy__.left_scalar_operator(other, '>='))


    def __radd__(self, other):
        """
        Adds a scalar value to the current array.
        Returned array has the same type as the array on the right hand side
        """
        with cython_context():
            return SArray(_proxy = self.__proxy__.right_scalar_operator(other, '+'))


    def __rsub__(self, other):
        """
        Subtracts a scalar value from the current array.
        Returned array has the same type as the array on the right hand side
        """
        with cython_context():
            return SArray(_proxy = self.__proxy__.right_scalar_operator(other, '-'))


    def __rmul__(self, other):
        """
        Multiplies a scalar value to the current array.
        Returned array has the same type as the array on the right hand side
        """
        with cython_context():
            return SArray(_proxy = self.__proxy__.right_scalar_operator(other, '*'))


    def __rdiv__(self, other):
        """
        Divides a scalar value by each element in the array
        Returned array has the same type as the array on the right hand side
        """
        with cython_context():
            return SArray(_proxy = self.__proxy__.right_scalar_operator(other, '/'))


    def __eq__(self, other):
        """
        If other is a scalar value, compares each element of the current array
        by the value, returning the new result. If other is an SArray, performs
        an element-wise comparison of the two arrays.
        """
        with cython_context():
            if type(other) is SArray:
                return SArray(_proxy = self.__proxy__.vector_operator(other.__proxy__, '=='))
            else:
                return SArray(_proxy = self.__proxy__.left_scalar_operator(other, '=='))


    def __ne__(self, other):
        """
        If other is a scalar value, compares each element of the current array
        by the value, returning the new result. If other is an SArray, performs
        an element-wise comparison of the two arrays.
        """
        with cython_context():
            if type(other) is SArray:
                return SArray(_proxy = self.__proxy__.vector_operator(other.__proxy__, '!='))
            else:
                return SArray(_proxy = self.__proxy__.left_scalar_operator(other, '!='))


    def __and__(self, other):
        """
        Perform a logical element-wise 'and' against another SArray.
        """
        if type(other) is SArray:
            with cython_context():
                return SArray(_proxy = self.__proxy__.vector_operator(other.__proxy__, '&'))
        else:
            raise TypeError("SArray can only perform logical and against another SArray")


    def __or__(self, other):
        """
        Perform a logical element-wise 'or' against another SArray.
        """
        if type(other) is SArray:
            with cython_context():
                return SArray(_proxy = self.__proxy__.vector_operator(other.__proxy__, '|'))
        else:
            raise TypeError("SArray can only perform logical or against another SArray")


    def __getitem__(self, other):
        """
        If the key is an SArray of identical length, this function performs a
        logical filter: i.e. it subselects all the elements in this array
        where the corresponding value in the other array evaluates to true.
        If the key is an integer this returns a single row of
        the SArray. If the key is a slice, this returns an SArray with the
        sliced rows. See the GraphLab Create User Guide for usage examples.
        """
        if type(other) is SArray:
            if len(other) != len(self):
                raise IndexError("Cannot perform logical indexing on arrays of different length.")
            with cython_context():
                return SArray(_proxy = self.__proxy__.logical_filter(other.__proxy__))
        elif type(other) is int:
            if other < 0:
                other = len(self) + other
            if other >= len(self):
                raise IndexError("SFrame index out of range")
            return list(SArray(_proxy = self.__proxy__.copy_range(other, 1, other+1)))[0]
        elif type(other) is slice:
            start = other.start
            stop = other.stop
            step = other.step
            if start is None:
                start = 0
            if stop is None:
                stop = len(self)
            if step is None:
                step = 1
            # handle negative indices
            if start < 0:
                start = len(self) + start
            if stop < 0:
                stop = len(self) + stop
            return SArray(_proxy = self.__proxy__.copy_range(start, step, stop))
        else:
            raise IndexError("Invalid type to use for indexing")

    def __materialize__(self):
        """
        For a SArray that is lazily evaluated, force persist this sarray
        to disk, committing all lazy evaluated operations.
        """
        with cython_context():
            self.__proxy__.materialize()

    def __is_materialized__(self):
        """
        Returns whether or not the sarray has been materialized.
        """
        return self.__proxy__.is_materialized()

    def size(self):
        """
        The size of the SArray.
        """
        return self.__proxy__.size()

    def dtype(self):
        """
        The data type of the SArray.

        Returns
        -------
        out : type
            The type of the sarray.

        Examples
        --------
        >>> sa = gl.SArray(["The quick brown fox jumps over the lazy dog."])
        >>> sa.dtype
        str
        >>> sa = gl.SArray(range(10))
        >>> sa.dtype
        int
        """
        return self.__proxy__.dtype()

    def head(self, n=10):
        """
        Returns an SArray which contains the first n rows of this SArray.

        Parameters
        ----------
        n : int
            The number of rows to fetch.

        Returns
        -------
        out : SArray
            A new SArray which contains the first n rows of the current SArray.

        Examples
        --------
        >>> gl.SArray(range(10)).head(5)
        dtype: int
        Rows: 5
        [0, 1, 2, 3, 4]
        """
        return SArray(_proxy=self.__proxy__.head(n))

    def vector_slice(self, start, end=None):
        """
        If this SArray contains vectors or recursive types, this returns a new SArray
        containing each individual vector sliced, between start and end, exclusive.

        Parameters
        ----------
        start : int
            The start position of the slice.

        end : int, optional.
            The end position of the slice. Note that the end position
            is NOT included in the slice. Thus a g.vector_slice(1,3) will extract
            entries in position 1 and 2.

        Returns
        -------
        out : SArray
            Each individual vector sliced according to the arguments.

        Examples
        --------

        If g is a vector of floats:

        >>> g = SArray([[1,2,3],[2,3,4]])
        >>> g
        dtype: array
        Rows: 2
        [array('d', [1.0, 2.0, 3.0]), array('d', [2.0, 3.0, 4.0])]

        >>> g.vector_slice(0) # extracts the first element of each vector
        dtype: float
        Rows: 2
        [1.0, 2.0]

        >>> g.vector_slice(0, 2) # extracts the first two elements of each vector
        dtype: array.array
        Rows: 2
        [array('d', [1.0, 2.0]), array('d', [2.0, 3.0])]

        If a vector cannot be sliced, the result will be None:

        >>> g = SArray([[1],[1,2],[1,2,3]])
        >>> g
        dtype: array.array
        Rows: 3
        [array('d', [1.0]), array('d', [1.0, 2.0]), array('d', [1.0, 2.0, 3.0])]

        >>> g.vector_slice(2)
        dtype: float
        Rows: 3
        [None, None, 3.0]

        >>> g.vector_slice(0,2)
        dtype: list
        Rows: 3
        [None, array('d', [1.0, 2.0]), array('d', [1.0, 2.0])]

        If g is a vector of mixed types (float, int, str, array, list, etc.):

        >>> g = SArray([['a',1,1.0],['b',2,2.0]])
        >>> g
        dtype: list
        Rows: 2
        [['a', 1, 1.0], ['b', 2, 2.0]]

        >>> g.vector_slice(0) # extracts the first element of each vector
        dtype: list
        Rows: 2
        [['a'], ['b']]
        """
        if (self.dtype() != array.array) and (self.dtype() != list):
            raise RuntimeError("Only Vector type can be sliced")
        if end == None:
            end = start + 1

        with cython_context():
            return SArray(_proxy=self.__proxy__.vector_slice(start, end))

    def count_words(self, to_lower=True):
        """
        Count words in the SArray. Return an SArray of dictionary type where
        each element contains the word count for each word that appeared in the
        corresponding input element. The words are split on all whitespace and
        punctuation characters. Only works if this SArray is of string type.

        Parameters
        ----------
        to_lower : bool, optional
            If True, all words are converted to lower case before counting.

        Returns
        -------
        out : SArray
            The SArray of dictionary type, where each element contains the word
            count for each word that appeared in corresponding input element.

        See Also
        --------
        count_ngrams

        Examples
        --------
        >>> sa = graphlab.SArray(["The quick brown fox jumps.",
                                 "Word word WORD, word!!!word"])
        >>> sa.count_words()
        dtype: dict
        Rows: 2
        [{'quick': 1, 'brown': 1, 'jumps': 1, 'fox': 1, 'the': 1}, {'word': 5}]
        """
        if (self.dtype() != str):
            raise TypeError("Only SArray of string type is supported for counting bag of words")

        _mt._get_metric_tracker().track('sarray.count_words')

        # construct options, will extend over time
        options = dict()
        options["to_lower"] = to_lower == True


        with cython_context():
            return SArray(_proxy=self.__proxy__.count_bag_of_words(options))

    def count_ngrams(self, n=2, method="word", to_lower=True, ignore_space=True):
        """
        Return an SArray of ``dict`` type where each element contains the count
        for each of the n-grams that appear in the corresponding input element.
        The n-grams can be specified to be either character n-grams or word
        n-grams.  The input SArray must contain strings.

        Parameters
        ----------
        n : int, optional
            The number of words in each n-gram. An ``n`` value of 1 returns word
            counts.

        method : {'word', 'character'}, optional
            If "word", the function performs a count of word n-grams. If
            "character", does a character n-gram count.

        to_lower : bool, optional
            If True, all words are converted to lower case before counting.

        ignore_space : bool, optional
            If method is "character", indicates if spaces between words are
            counted as part of the n-gram. For instance, with the input SArray
            element of "fun games", if this parameter is set to False one
            tri-gram would be 'n g'. If ``ignore_space`` is set to True, there
            would be no such tri-gram (there would still be 'nga'). This
            parameter has no effect if the method is set to "word".

        Returns
        -------
        out : SArray
            An SArray of dictionary type, where each key is the n-gram string
            and each value is its count.

        See Also
        --------
        count_words

        Notes
        -----
        - Ignoring case (with ``to_lower``) involves a full string copy of the
          SArray data. To increase speed for large documents, set ``to_lower``
          to False.

        - Punctuation and spaces are both delimiters when counting word n-grams.
          When counting character n-grams, punctuation is always ignored.

        References
        ----------
        - `N-gram wikipedia article <http://en.wikipedia.org/wiki/N-gram>`_

        Examples
        --------
        Counting word n-grams:

        >>> from graphlab import SArray
        >>> sa = SArray(['I like big dogs. I LIKE BIG DOGS.'])
        >>> sa.count_ngrams(3)
        dtype: dict
        Rows: 1
        [{'big dogs i': 1, 'like big dogs': 2, 'dogs i like': 1, 'i like big': 2}]

        Counting character n-grams:

        >>> sa = SArray(['Fun. Is. Fun'])
        >>> sa.count_ngrams(3, "character")
        dtype: dict
        Rows: 1
        {'fun': 2, 'nis': 1, 'sfu': 1, 'isf': 1, 'uni': 1}]
        """
        if (self.dtype() != str):
            raise TypeError("Only SArray of string type is supported for counting n-grams")

        if (type(n) != int):
            raise TypeError("Input 'n' must be of type int")

        if (n < 1):
            raise ValueError("Input 'n' must be greater than 0")

        if (n > 5):
            warnings.warn("It is unusual for n-grams to be of size larger than 5.")

        _mt._get_metric_tracker().track('sarray.count_ngrams', properties={'n':n, 'method':method})

        # construct options, will extend over time
        options = dict()
        options["to_lower"] = to_lower == True
        options["ignore_space"] = ignore_space == True


        if method == "word":
            with cython_context():
                return SArray(_proxy=self.__proxy__.count_ngrams(n, options ))
        elif method == "character" :
            with cython_context():
                return SArray(_proxy=self.__proxy__.count_character_ngrams(n, options ))
        else:
            raise ValueError("Invalid 'method' input  value. Please input either 'word' or 'character' ")

    def dict_trim_by_keys(self, keys, exclude=True):
        """
        Filter an SArray of dictionary type by the given keys. By default, all
        keys that are in the provided list in ``keys`` are *excluded* from the
        returned SArray.

        Parameters
        ----------
        keys : list
            A collection of keys to trim down the elements in the SArray.

        exclude : bool, optional
            If True, all keys that are in the input key list are removed. If
            False, only keys that are in the input key list are retained.

        Returns
        -------
        out : SArray
            A SArray of dictionary type, with each dictionary element trimmed
            according to the input criteria.

        See Also
        --------
        dict_trim_by_values

        Examples
        --------
        >>> sa = graphlab.SArray([{"this":1, "is":1, "dog":2},
                                  {"this": 2, "are": 2, "cat": 1}])
        >>> sa.dict_trim_by_keys(["this", "is", "and", "are"], exclude=True)
        dtype: dict
        Rows: 2
        [{'dog': 2}, {'cat': 1}]
        """
        if isinstance(keys, str) or (not hasattr(keys, "__iter__")):
            keys = [keys]

        _mt._get_metric_tracker().track('sarray.dict_trim_by_keys')

        with cython_context():
            return SArray(_proxy=self.__proxy__.dict_trim_by_keys(keys, exclude))

    def dict_trim_by_values(self, lower=None, upper=None):
        """
        Filter dictionary values to a given range (inclusive). Trimming is only
        performed on values which can be compared to the bound values. Fails on
        SArrays whose data type is not ``dict``.

        Parameters
        ----------
        lower : int or long or float, optional
            The lowest dictionary value that would be retained in the result. If
            not given, lower bound is not applied.

        upper : int or long or float, optional
            The highest dictionary value that would be retained in the result.
            If not given, upper bound is not applied.

        Returns
        -------
        out : SArray
            An SArray of dictionary type, with each dict element trimmed
            according to the input criteria.

        See Also
        --------
        dict_trim_by_keys

        Examples
        --------
        >>> sa = graphlab.SArray([{"this":1, "is":5, "dog":7},
                                  {"this": 2, "are": 1, "cat": 5}])
        >>> sa.dict_trim_by_values(2,5)
        dtype: dict
        Rows: 2
        [{'is': 5}, {'this': 2, 'cat': 5}]

        >>> sa.dict_trim_by_values(upper=5)
        dtype: dict
        Rows: 2
        [{'this': 1, 'is': 5}, {'this': 2, 'are': 1, 'cat': 5}]
        """

        if None != lower and (not is_numeric_type(type(lower))):
            raise TypeError("lower bound has to be a numeric value")

        if None != upper and (not is_numeric_type(type(upper))):
            raise TypeError("upper bound has to be a numeric value")

        _mt._get_metric_tracker().track('sarray.dict_trim_by_values')

        with cython_context():
            return SArray(_proxy=self.__proxy__.dict_trim_by_values(lower, upper))

    def dict_keys(self):
        """
        Create an SArray that contains all the keys from each dictionary
        element as a list. Fails on SArrays whose data type is not ``dict``.

        Returns
        -------
        out : SArray
            A SArray of list type, where each element is a list of keys
            from the input SArray element.

        See Also
        --------
        dict_values

        Examples
        ---------
        >>> sa = graphlab.SArray([{"this":1, "is":5, "dog":7},
                                  {"this": 2, "are": 1, "cat": 5}])
        >>> sa.dict_keys()
        dtype: list
        Rows: 2
        [['this', 'is', 'dog'], ['this', 'are', 'cat']]
        """
        _mt._get_metric_tracker().track('sarray.dict_keys')

        with cython_context():
            return SArray(_proxy=self.__proxy__.dict_keys())

    def dict_values(self):
        """
        Create an SArray that contains all the values from each dictionary
        element as a list. Fails on SArrays whose data type is not ``dict``.

        Returns
        -------
        out : SArray
            A SArray of list type, where each element is a list of values
            from the input SArray element.

        See Also
        --------
        dict_keys

        Examples
        --------
        >>> sa = graphlab.SArray([{"this":1, "is":5, "dog":7},
                                 {"this": 2, "are": 1, "cat": 5}])
        >>> sa.dict_values()
        dtype: list
        Rows: 2
        [[1, 5, 7], [2, 1, 5]]

        """
        _mt._get_metric_tracker().track('sarray.dict_values')

        with cython_context():
            return SArray(_proxy=self.__proxy__.dict_values())

    def dict_has_any_keys(self, keys):
        """
        Create a boolean SArray by checking the keys of an SArray of
        dictionaries. An element of the output SArray is True if the
        corresponding input element's dictionary has any of the given keys.
        Fails on SArrays whose data type is not ``dict``.

        Parameters
        ----------
        keys : list
            A list of key values to check each dictionary against.

        Returns
        -------
        out : SArray
            A SArray of int type, where each element indicates whether the
            input SArray element contains any key in the input list.

        See Also
        --------
        dict_has_all_keys

        Examples
        --------
        >>> sa = graphlab.SArray([{"this":1, "is":5, "dog":7}, {"animal":1},
                                 {"this": 2, "are": 1, "cat": 5}])
        >>> sa.dict_has_any_keys(["is", "this", "are"])
        dtype: int
        Rows: 3
        [1, 1, 0]
        """
        if isinstance(keys, str) or (not hasattr(keys, "__iter__")):
            keys = [keys]

        _mt._get_metric_tracker().track('sarray.dict_has_any_keys')

        with cython_context():
            return SArray(_proxy=self.__proxy__.dict_has_any_keys(keys))

    def dict_has_all_keys(self, keys):
        """
        Create a boolean SArray by checking the keys of an SArray of
        dictionaries. An element of the output SArray is True if the
        corresponding input element's dictionary has all of the given keys.
        Fails on SArrays whose data type is not ``dict``.

        Parameters
        ----------
        keys : list
            A list of key values to check each dictionary against.

        Returns
        -------
        out : SArray
            A SArray of int type, where each element indicates whether the
            input SArray element contains all keys in the input list.

        See Also
        --------
        dict_has_any_keys

        Examples
        --------
        >>> sa = graphlab.SArray([{"this":1, "is":5, "dog":7},
                                 {"this": 2, "are": 1, "cat": 5}])
        >>> sa.dict_has_all_keys(["is", "this"])
        dtype: int
        Rows: 2
        [1, 0]
        """
        if isinstance(keys, str) or (not hasattr(keys, "__iter__")):
            keys = [keys]

        _mt._get_metric_tracker().track('sarray.dict_has_all_keys')

        with cython_context():
            return SArray(_proxy=self.__proxy__.dict_has_all_keys(keys))

    def apply(self, fn, dtype=None, skip_undefined=True, seed=None,
              _lua_translate=False):
        """
        apply(fn, dtype=None, skip_undefined=True, seed=None)

        Transform each element of the SArray by a given function. The result
        SArray is of type ``dtype``. ``fn`` should be a function that returns
        exactly one value which can be cast into the type specified by
        ``dtype``. If ``dtype`` is not specified, the first 100 elements of the
        SArray are used to make a guess about the data type.

        Parameters
        ----------
        fn : function
            The function to transform each element. Must return exactly one
            value which can be cast into the type specified by ``dtype``.

        dtype : {None, int, float, str, list, array.array, dict}, optional
            The data type of the new SArray. If ``None``, the first 100 elements
            of the array are used to guess the target data type.

        skip_undefined : bool, optional
            If True, will not apply ``fn`` to any undefined values.

        seed : int, optional
            Used as the seed if a random number generator is included in ``fn``.

        Returns
        -------
        out : SArray
            The SArray transformed by ``fn``. Each element of the SArray is of
            type ``dtype``.

        See Also
        --------
        SFrame.apply

        Examples
        --------
        >>> sa = graphlab.SArray([1,2,3])
        >>> sa.apply(lambda x: x*2)
        dtype: int
        Rows: 3
        [2, 4, 6]
        """
        if (type(fn) == str):
            fn = "LUA" + fn
            if dtype == None:
                raise TypeError("dtype must be specified for a lua function")
        else:
            assert inspect.isfunction(fn), "Input must be a function"

            dryrun = [fn(i) for i in self.head(100) if i is not None]
            import traceback
            if dtype == None:
                dtype = infer_type_of_list(dryrun)
            try:
                # try compilation
                if _lua_translate:
                    # its a function
                    print "Attempting Lua Translation"
                    import graphlab.Lua_Translator
                    import ast
                    import StringIO

                    def isalambda(v):
                        return isinstance(v, type(lambda: None)) and v.__name__ == '<lambda>'

                    output = StringIO.StringIO()
                    translator = graphlab.Lua_Translator.translator_NodeVisitor(output)
                    ast_node = None
                    try:
                        if not isalambda(fn):
                            ast_node = ast.parse(inspect.getsource(fn))
                            translator.rename_function[fn.__name__] = "__lambda__transfer__"
                    except:
                        pass

                    try:
                        if ast_node == None:
                            print "Cannot translate. Trying again from byte code decompilation"
                            ast_node = meta.decompiler.decompile_func(fn)
                            translator.rename_function[""] = "__lambda__transfer__"
                    except:
                        pass
                    if ast_node == None:
                        raise ValueError("Unable to get source of function")

                    ftype = graphlab.Lua_Translator.FunctionType()
                    selftype = self.dtype()
                    if selftype == list:
                        ftype.input_type = tuple([[]])
                    elif selftype == dict:
                        ftype.input_type = tuple([{}])
                    elif selftype == array.array:
                        ftype.input_type = tuple([[float]])
                    else:
                        ftype.input_type = tuple([selftype])
                    translator.function_known_types["__lambda__transfer__"] = ftype
                    translator.translate_ast(ast_node)
                    print "Lua Translation Success"
                    print output.getvalue()
                    fn = "LUA" + output.getvalue()
            except Exception as e:
                print traceback.format_exc()
                print "Lua Translation Failed"
                print e
            except:
                print traceback.format_exc()
                print "Lua Translation Failed"

        if not seed:
            seed = time.time()

        _mt._get_metric_tracker().track('sarray.apply')

        with cython_context():
            return SArray(_proxy=self.__proxy__.transform(fn, dtype, skip_undefined, seed))


    def filter(self, fn, skip_undefined=True, seed=None):
        """
        Filter this SArray by a function.

        Returns a new SArray filtered by this SArray.  If `fn` evaluates an
        element to true, this element is copied to the new SArray. If not, it
        isn't. Throws an exception if the return type of `fn` is not castable
        to a boolean value.

        Parameters
        ----------
        fn : function
            Function that filters the SArray. Must evaluate to bool or int.

        skip_undefined : bool, optional
            If True, will not apply fn to any undefined values.

        seed : int, optional
            Used as the seed if a random number generator is included in fn.

        Returns
        -------
        out : SArray
            The SArray filtered by fn. Each element of the SArray is of
            type int.

        Examples
        --------
        >>> sa = graphlab.SArray([1,2,3])
        >>> sa.filter(lambda x: x < 3)
        dtype: int
        Rows: 2
        [1, 2]
        """
        assert inspect.isfunction(fn), "Input must be a function"
        if not seed:
            seed = time.time()

        _mt._get_metric_tracker().track('sarray.filter')

        with cython_context():
            return SArray(_proxy=self.__proxy__.filter(fn, skip_undefined, seed))


    def sample(self, fraction, seed=None):
        """
        Create an SArray which contains a subsample of the current SArray.

        Parameters
        ----------
        fraction : float
            The fraction of the rows to fetch. Must be between 0 and 1.

        seed : int
            The random seed for the random number generator.

        Returns
        -------
        out : SArray
            The new SArray which contains the subsampled rows.

        Examples
        --------
        >>> sa = graphlab.SArray(range(10))
        >>> sa.sample(.3)
        dtype: int
        Rows: 3
        [2, 6, 9]
        """
        if (fraction > 1 or fraction < 0):
            raise ValueError('Invalid sampling rate: ' + str(fraction))
        if (self.size() == 0):
            return SArray()
        if not seed:
            seed = time.time()

        _mt._get_metric_tracker().track('sarray.sample')

        with cython_context():
            return SArray(_proxy=self.__proxy__.sample(fraction, seed))

    def _save_as_text(self, url):
        """
        Save the SArray to disk as text file.
        """
        raise NotImplementedError


    def all(self):
        """
        Return True if every element of the SArray evaluates to False. For
        numeric SArrays zeros and missing values (``None``) evaluate to False,
        while all non-zero, non-missing values evaluate to True. For string,
        list, and dictionary SArrays, empty values (zero length strings, lists
        or dictionaries) or missing values (``None``) evaluate to False. All
        other values evaluate to True.

        Returns True on an empty SArray.

        Returns
        -------
        out : bool

        See Also
        --------
        any

        Examples
        --------
        >>> graphlab.SArray([1, None]).all()
        False
        >>> graphlab.SArray([1, 0]).all()
        False
        >>> graphlab.SArray([1, 2]).all()
        True
        >>> graphlab.SArray(["hello", "world"]).all()
        True
        >>> graphlab.SArray(["hello", ""]).all()
        False
        >>> graphlab.SArray([]).all()
        True
        """
        with cython_context():
            return self.__proxy__.all()


    def any(self):
        """
        Return True if any element of the SArray evaluates to True. For numeric
        SArrays any non-zero value evaluates to True. For string, list, and
        dictionary SArrays, any element of non-zero length evaluates to True.

        Returns False on an empty SArray.

        Returns
        -------
        out : bool

        See Also
        --------
        all

        Examples
        --------
        >>> graphlab.SArray([1, None]).any()
        True
        >>> graphlab.SArray([1, 0]).any()
        True
        >>> graphlab.SArray([0, 0]).any()
        False
        >>> graphlab.SArray(["hello", "world"]).any()
        True
        >>> graphlab.SArray(["hello", ""]).any()
        True
        >>> graphlab.SArray(["", ""]).any()
        False
        >>> graphlab.SArray([]).any()
        False
        """
        with cython_context():
            return self.__proxy__.any()


    def max(self):
        """
        Get maximum numeric value in SArray.

        Returns None on an empty SArray. Raises an exception if called on an
        SArray with non-numeric type.

        Returns
        -------
        out : type of SArray
            Maximum value of SArray

        See Also
        --------
        min

        Examples
        --------
        >>> graphlab.SArray([14, 62, 83, 72, 77, 96, 5, 25, 69, 66]).max()
        96
        """
        with cython_context():
            return self.__proxy__.max()


    def min(self):
        """
        Get minimum numeric value in SArray.

        Returns None on an empty SArray. Raises an exception if called on an
        SArray with non-numeric type.

        Returns
        -------
        out : type of SArray
            Minimum value of SArray

        See Also
        --------
        max

        Examples
        --------
        >>> graphlab.SArray([14, 62, 83, 72, 77, 96, 5, 25, 69, 66]).min()

        """
        with cython_context():
            return self.__proxy__.min()


    def sum(self):
        """
        Sum of all values in this SArray.

        Raises an exception if called on an SArray of strings, lists, or
        dictionaries. If the SArray contains numeric arrays (array.array) and
        all the arrays are the same length, the sum over all the arrays will be
        returned. Returns None on an empty SArray. For large values, this may
        overflow without warning.

        Returns
        -------
        out : type of SArray
            Sum of all values in SArray
        """
        with cython_context():
            return self.__proxy__.sum()


    def mean(self):
        """
        Mean of all the values in the SArray.

        Returns None on an empty SArray. Raises an exception if called on an
        SArray with non-numeric type.

        Returns
        -------
        out : float
            Mean of all values in SArray
        """
        with cython_context():
            return self.__proxy__.mean()


    def std(self, ddof=0):
        """
        Standard deviation of all the values in the SArray.

        Returns None on an empty SArray. Raises an exception if called on an
        SArray with non-numeric type or if `ddof` >= length of SArray.

        Parameters
        ----------
        ddof : int, optional
            "delta degrees of freedom" in the variance calculation.

        Returns
        -------
        out : float
            The standard deviation of all the values.
        """
        with cython_context():
            return self.__proxy__.std(ddof)


    def var(self, ddof=0):
        """
        Variance of all the values in the SArray.

        Returns None on an empty SArray. Raises an exception if called on an
        SArray with non-numeric type or if `ddof` >= length of SArray.

        Parameters
        ----------
        ddof : int, optional
            "delta degrees of freedom" in the variance calculation.

        Returns
        -------
        out : float
            Variance of all values in SArray.
        """
        with cython_context():
            return self.__proxy__.var(ddof)

    def num_missing(self):
        """
        Number of missing elements in the SArray.

        Returns
        -------
        out : int
            Number of missing values.
        """
        with cython_context():
            return self.__proxy__.num_missing()

    def nnz(self):
        """
        Number of non-zero elements in the SArray.

        Returns
        -------
        out : int
            Number of non-zero elements.
        """
        with cython_context():
            return self.__proxy__.nnz()

    def astype(self, dtype, undefined_on_failure=False):
        """
        Create a new SArray with all values cast to the given type. Throws an
        exception if the types are not castable to the given type.

        Parameters
        ----------
        dtype : {int, float, str, list, array.array, dict}
            The type to cast the elements to in SArray

        undefined_on_failure: bool, optional
            If set to True, runtime cast failures will be emitted as missing
            values rather than failing.

        Returns
        -------
        out : SArray [dtype]
            The SArray converted to the type ``dtype``.

        Notes
        -----
        - The string parsing techniques used to handle conversion to dictionary
          and list types are quite generic and permit a variety of interesting
          formats to be interpreted. For instance, a JSON string can usually be
          interpreted as a list or a dictionary type. See the examples below.

        Examples
        --------
        >>> sa = graphlab.SArray(['1','2','3','4'])
        >>> sa.astype(int)
        dtype: int
        Rows: 4
        [1, 2, 3, 4]

        Given an SArray of strings that look like dicts, convert to a dictionary
        type:

        >>> sa = graphlab.SArray(['{1:2 3:4}', '{a:b c:d}'])
        >>> sa.astype(dict)
        dtype: dict
        Rows: 2
        [{1: 2, 3: 4}, {'a': 'b', 'c': 'd'}]
        """

        _mt._get_metric_tracker().track('sarray.astype.%s' % str(dtype.__name__))

        with cython_context():
            return SArray(_proxy=self.__proxy__.astype(dtype, undefined_on_failure))

    def clip(self, lower=float('nan'), upper=float('nan')):
        """
        Create a new SArray with each value clipped to be within the given
        bounds.

        In this case, "clipped" means that values below the lower bound will be
        set to the lower bound value. Values above the upper bound will be set
        to the upper bound value. This function can operate on SArrays of
        numeric type as well as array type, in which case each individual
        element in each array is clipped. By default ``lower`` and ``upper`` are
        set to ``float('nan')`` which indicates the respective bound should be
        ignored. The method fails if invoked on an SArray of non-numeric type.

        Parameters
        ----------
        lower : int, optional
            The lower bound used to clip. Ignored if equal to ``float('nan')``
            (the default).

        upper : int, optional
            The upper bound used to clip. Ignored if equal to ``float('nan')``
            (the default).

        Returns
        -------
        out : SArray

        See Also
        --------
        clip_lower, clip_upper

        Examples
        --------
        >>> sa = graphlab.SArray([1,2,3])
        >>> sa.clip(2,2)
        dtype: int
        Rows: 3
        [2, 2, 2]
        """
        with cython_context():
            return SArray(_proxy=self.__proxy__.clip(lower, upper))

    def clip_lower(self, threshold):
        """
        Create new SArray with all values clipped to the given lower bound. This
        function can operate on numeric arrays, as well as vector arrays, in
        which case each individual element in each vector is clipped. Throws an
        exception if the SArray is empty or the types are non-numeric.

        Parameters
        ----------
        threshold : float
            The lower bound used to clip values.

        Returns
        -------
        out : SArray

        See Also
        --------
        clip, clip_upper

        Examples
        --------
        >>> sa = graphlab.SArray([1,2,3])
        >>> sa.clip_lower(2)
        dtype: int
        Rows: 3
        [2, 2, 3]
        """
        with cython_context():
            return SArray(_proxy=self.__proxy__.clip(threshold, float('nan')))


    def clip_upper(self, threshold):
        """
        Create new SArray with all values clipped to the given upper bound. This
        function can operate on numeric arrays, as well as vector arrays, in
        which case each individual element in each vector is clipped.

        Parameters
        ----------
        threshold : float
            The upper bound used to clip values.

        Returns
        -------
        out : SArray

        See Also
        --------
        clip, clip_lower

        Examples
        --------
        >>> sa = graphlab.SArray([1,2,3])
        >>> sa.clip_upper(2)
        dtype: int
        Rows: 3
        [1, 2, 2]
        """
        with cython_context():
            return SArray(_proxy=self.__proxy__.clip(float('nan'), threshold))

    def tail(self, n=10):
        """
        Get an SArray that contains the last n elements in the SArray.

        Parameters
        ----------
        n : int
            The number of elements to fetch

        Returns
        -------
        out : SArray
            A new SArray which contains the last n rows of the current SArray.
        """
        with cython_context():
            return SArray(_proxy=self.__proxy__.tail(n))


    def dropna(self):
        """
        Create new SArray containing only the non-missing values of the
        SArray.

        A missing value shows up in an SArray as 'None'.  This will also drop
        float('nan').

        Returns
        -------
        out : SArray
            The new SArray with missing values removed.
        """

        _mt._get_metric_tracker().track('sarray.dropna')

        with cython_context():
            return SArray(_proxy = self.__proxy__.drop_missing_values())

    def fillna(self, value):
        """
        Create new SArray with all missing values (None or NaN) filled in
        with the given value.

        The size of the new SArray will be the same as the original SArray. If
        the given value is not the same type as the values in the SArray,
        `fillna` will attempt to convert the value to the original SArray's
        type. If this fails, an error will be raised.

        Parameters
        ----------
        value : type convertible to SArray's type
            The value used to replace all missing values

        Returns
        -------
        out : SArray
            A new SArray with all missing values filled
        """
        _mt._get_metric_tracker().track('sarray.fillna')

        with cython_context():
            return SArray(_proxy = self.__proxy__.fill_missing_values(value))

    def topk_index(self, topk=10, reverse=False):
        """
        Create an SArray indicating which elements are in the top k.

        Entries are '1' if the corresponding element in the current SArray is a
        part of the top k elements, and '0' if that corresponding element is
        not. Order is descending by default.

        Parameters
        ----------
        topk : int
            The number of elements to determine if 'top'

        reverse: bool
            If True, return the topk elements in ascending order

        Returns
        -------
        out : SArray (of type int)

        Notes
        -----
        This is used internally by SFrame's topk function.
        """
        with cython_context():
            return SArray(_proxy = self.__proxy__.topk_index(topk, reverse))

    def sketch_summary(self, background=False, sub_sketch_keys=None):
        """
        Summary statistics that can be calculated with one pass over the SArray.

        Returns a graphlab.Sketch object which can be further queried for many
        descriptive statistics over this SArray. Many of the statistics are
        approximate. See the :class:`~graphlab.Sketch` documentation for more
        detail.

        Parameters
        ----------
        background : boolean, optional
          If True, the sketch construction will return immediately and the
          sketch will be constructed in the background. While this is going on,
          the sketch can be queried incrementally, but at a performance penalty.
          Defaults to False.

        sub_sketch_keys: int | str | list of int | list of str, optional
            For SArray of dict type, also constructs sketches for a given set of keys,
            For SArray of array type, also constructs sketches for the given indexes.
            The sub sketches may be queried using:
                 :py:func:`~graphlab.Sketch.element_sub_sketch()`
            Defaults to None in which case no subsketches will be constructed.

        Returns
        -------
        out : Sketch
            Sketch object that contains descriptive statistics for this SArray.
            Many of the statistics are approximate.
        """
        from graphlab.data_structures.sketch import Sketch
        if (type(background) != bool):
            raise TypeError("'background' parameter has to be a boolean value")
        if (sub_sketch_keys != None):
            if (self.dtype() != dict and self.dtype() != array.array):
                raise TypeError("sub_sketch_keys is only supported for SArray of dictionary or array type")
            if not hasattr(sub_sketch_keys, "__iter__"):
                sub_sketch_keys = [sub_sketch_keys]
            value_types = set([type(i) for i in sub_sketch_keys])
            if (len(value_types) != 1):
                raise ValueError("sub_sketch_keys member values need to have the same type.")
            value_type = value_types.pop();
            if (self.dtype() == dict and value_type != str):
                raise TypeError("Only string value(s) can be passed to sub_sketch_keys for SArray of dictionary type. "+
                    "For dictionary types, sketch summary is computed by casting keys to string values.")
            if (self.dtype() == array.array and value_type != int):
                raise TypeError("Only int value(s) can be passed to sub_sketch_keys for SArray of array type")
        else:
            sub_sketch_keys = list()

        _mt._get_metric_tracker().track('sarray.sketch_summary')
        return Sketch(self, background, sub_sketch_keys = sub_sketch_keys)

    def append(self, other):
        """
        Append an SArray to the current SArray. Creates a new SArray with the
        rows from both SArrays. Both SArrays must be of the same type.

        Parameters
        ----------
        other : SArray
            Another SArray whose rows are appended to current SArray.

        Returns
        -------
        out : SArray
            A new SArray that contains rows from both SArrays, with rows from
            the ``other`` SArray coming after all rows from the current SArray.

        See Also
        --------
        SFrame.append

        Examples
        --------
        >>> sa = graphlab.SArray([1, 2, 3])
        >>> sa2 = graphlab.SArray([4, 5, 6])
        >>> sa.append(sa2)
        dtype: int
        Rows: 6
        [1, 2, 3, 4, 5, 6]
        """
        _mt._get_metric_tracker().track('sarray.append')
        if type(other) is not SArray:
            raise RuntimeError("SArray append can only work with SArray")

        if self.dtype() != other.dtype():
            raise RuntimeError("Data types in both SArrays have to be the same")

        with cython_context():
            return SArray(_proxy = self.__proxy__.append(other.__proxy__))

    def unique(self):
        """
        Get all unique values in the current SArray.

        Raises a TypeError if the SArray is of dictionary type. Will not
        necessarily preserve the order of the given SArray in the new SArray.


        Returns
        -------
        out : SArray
            A new SArray that contains the unique values of the current SArray.

        See Also
        --------
        SFrame.unique
        """
        _mt._get_metric_tracker().track('sarray.unique')
        tmp_sf = gl.SFrame()
        tmp_sf.add_column(self, 'X1')

        res = tmp_sf.groupby('X1',{})

        return SArray(_proxy=res['X1'].__proxy__)

    @graphlab.canvas.inspect.find_vars
    def show(self):
        """
        show()
        Visualize the SArray with GraphLab Canvas. This function starts Canvas
        if it is not already running. If the SArray has already been plotted,
        this function will update the plot.

        Returns
        -------
        view : graphlab.canvas.view.View
            An object representing the GraphLab Canvas view

        See Also
        --------
        canvas

        Examples
        --------
        Suppose 'sa' is an SArray, we can view it in GraphLab Canvas using:

        >>> sa.show()
        """
        return graphlab.canvas.show(graphlab.canvas.views.sarray.SArrayView(self))

    def item_length(self):
        """
        Length of each element in the current SArray.

        Only works on SArrays of dict, array, or list type. If a given element
        is a missing value, then the output elements is also a missing value.
        This function is equivalent to the following but more performant:

            sa_item_len =  sa.apply(lambda x: len(x) if x is not None else None)

        Returns
        -------
        out_sf : SArray
            A new SArray, each element in the SArray is the len of the corresponding
            items in original SArray.

        Examples
        --------
        >>> sa = SArray([
        ...  {"is_restaurant": 1, "is_electronics": 0},
        ...  {"is_restaurant": 1, "is_retail": 1, "is_electronics": 0},
        ...  {"is_restaurant": 0, "is_retail": 1, "is_electronics": 0},
        ...  {"is_restaurant": 0},
        ...  {"is_restaurant": 1, "is_electronics": 1},
        ...  None])
        >>> sa.item_length()
        dtype: int
        Rows: 6
        [2, 3, 3, 1, 2, None]
        """
        if (self.dtype() not in [list, dict, array.array]):
            raise TypeError("item_length() is only applicable for SArray of type list, dict and array.")

        _mt._get_metric_tracker().track('sarray.item_length')

        with cython_context():
            return SArray(_proxy = self.__proxy__.item_length())

    def unpack(self, column_name_prefix = "X", column_types=None, na_value=None, limit=None):
        """
        Convert an SArray of list, array, or dict type to an SFrame with
        multiple columns.

        `unpack` expands an SArray using the values of each list/array/dict as
        elements in a new SFrame of multiple columns. For example, an SArray of
        lists each of length 4 will be expanded into an SFrame of 4 columns,
        one for each list element. An SArray of lists/arrays of varying size
        will be expand to a number of columns equal to the longest list/array.
        An SArray of dictionaries will be expanded into as many columns as
        there are keys.

        When unpacking an SArray of list or array type, new columns are named:
        `column_name_prefix`.0, `column_name_prefix`.1, etc. If unpacking a
        column of dict type, unpacked columns are named
        `column_name_prefix`.key1, `column_name_prefix`.key2, etc.

        When unpacking an SArray of list or dictionary types, missing values in
        the original element remain as missing values in the resultant columns.
        If the `na_value` parameter is specified, all values equal to this
        given value are also replaced with missing values. In an SArray of
        array.array type, NaN is interpreted as a missing value.

        :py:func:`graphlab.SFrame.pack_columns()` is the reverse effect of unpack

        Parameters
        ----------
        column_name_prefix: str, optional
            If provided, unpacked column names would start with the given prefix.

        column_types: list[type], optional
            Column types for the unpacked columns. If not provided, column
            types are automatically inferred from first 100 rows. Defaults to
            None.

        na_value: optional
            Convert all values that are equal to `na_value` to
            missing value if specified.

        limit: list, optional
            Limits the set of list/array/dict keys to unpack.
            For list/array SArrays, 'limit' must contain integer indices.
            For dict SArray, 'limit' must contain dictionary keys.

        Returns
        -------
        out : SFrame
            A new SFrame that contains all unpacked columns

        Examples
        --------
        To unpack a dict SArray

        >>> sa = SArray([{ 'word': 'a',     'count': 1},
        ...              { 'word': 'cat',   'count': 2},
        ...              { 'word': 'is',    'count': 3},
        ...              { 'word': 'coming','count': 4}])

        Normal case of unpacking SArray of type dict:

        >>> sa.unpack(column_name_prefix=None)
        Columns:
            count   int
            word    str
        <BLANKLINE>
        Rows: 4
        <BLANKLINE>
        Data:
        +-------+--------+
        | count |  word  |
        +-------+--------+
        |   1   |   a    |
        |   2   |  cat   |
        |   3   |   is   |
        |   4   | coming |
        +-------+--------+
        [4 rows x 2 columns]
        <BLANKLINE>

        Unpack only keys with 'word':

        >>> sa.unpack(limit=['word'])
        Columns:
            X.word  str
        <BLANKLINE>
        Rows: 4
        <BLANKLINE>
        Data:
        +--------+
        | X.word |
        +--------+
        |   a    |
        |  cat   |
        |   is   |
        | coming |
        +--------+
        [4 rows x 1 columns]
        <BLANKLINE>

        >>> sa2 = SArray([
        ...               [1, 0, 1],
        ...               [1, 1, 1],
        ...               [0, 1]])

        Convert all zeros to missing values:

        >>> sa2.unpack(column_types=[int, int, int], na_value=0)
        Columns:
            X.0     int
            X.1     int
            X.2     int
        <BLANKLINE>
        Rows: 3
        <BLANKLINE>
        Data:
        +------+------+------+
        | X.0  | X.1  | X.2  |
        +------+------+------+
        |  1   | None |  1   |
        |  1   |  1   |  1   |
        | None |  1   | None |
        +------+------+------+
        [3 rows x 3 columns]
        <BLANKLINE>
        """
        if self.dtype() not in [dict, array.array, list]:
            raise TypeError("Only SArray of dict/list/array type supports unpack")

        if column_name_prefix == None:
            column_name_prefix = ""
        if type(column_name_prefix) != str:
            raise TypeError("'column_name_prefix' must be a string")

        # validdate 'limit'
        if limit != None:
            if (not hasattr(limit, '__iter__')):
                raise TypeError("'limit' must be a list");

            name_types = set([type(i) for i in limit])
            if (len(name_types) != 1):
                raise TypeError("'limit' contains values that are different types")

            # limit value should be numeric if unpacking sarray.array value
            if (self.dtype() != dict) and (name_types.pop() != int):
                raise TypeError("'limit' must contain integer values.")

            if len(set(limit)) != len(limit):
                raise ValueError("'limit' contains duplicate values")

        if (column_types != None):
            if not hasattr(column_types, '__iter__'):
                raise TypeError("column_types must be a list");

            for column_type in column_types:
                if (column_type not in (int, float, str, list, dict, array.array)):
                    raise TypeError("column_types contains unsupported types. Supported types are ['float', 'int', 'list', 'dict', 'str', 'array.array']")

            if limit != None:
                if len(limit) != len(column_types):
                    raise ValueError("limit and column_types do not have the same length")
            elif self.dtype() == dict:
                raise ValueError("if 'column_types' is given, 'limit' has to be provided to unpack dict type.")
            else:
                limit = range(len(column_types))

        else:
            head_rows = self.head(100).dropna()
            lengths = [len(i) for i in head_rows]
            if len(lengths) == 0 or max(lengths) == 0:
                raise RuntimeError("Cannot infer number of items from the SArray, SArray may be empty. please explicitly provide column types")

            # infer column types for dict type at server side, for list and array, infer from client side
            if self.dtype() != dict:
                length = max(lengths)
                if limit == None:
                    limit = range(length)
                else:
                    # adjust the length
                    length = len(limit)

                if self.dtype() == array.array:
                    column_types = [float for i in range(length)]
                else:
                    column_types = list()
                    for i in limit:
                        t = [(x[i] if ((x is not None) and len(x) > i) else None) for x in head_rows]
                        column_types.append(infer_type_of_list(t))

        _mt._get_metric_tracker().track('sarray.unpack')

        with cython_context():
            if (self.dtype() == dict and column_types == None):
                limit = limit if limit != None else []
                return gl.SFrame(_proxy=self.__proxy__.unpack_dict(column_name_prefix, limit, na_value))
            else:
                return gl.SFrame(_proxy=self.__proxy__.unpack(column_name_prefix, limit, column_types, na_value))

    def sort(self, ascending=True):
        """
        Sort all values in this SArray.

        Sort only works for sarray of type str, int and float, otherwise TypeError
        will be raised. Creates a new, sorted SArray.

        Parameters
        ----------
        ascending: boolean, optional
           If true, the sarray values are sorted in ascending order, otherwise,
           descending order.

        Returns
        -------
        out: SArray

        Examples
        --------
        >>> sa = SArray([3,2,1])
        >>> sa.sort()
        dtype: int
        Rows: 3
        [1, 2, 3]
        """
        if self.dtype() not in (int, float, str):
            raise TypeError("Only sarray with type (int, float, str) can be sorted")
        sf = gl.SFrame()
        sf['a'] = self
        return sf.sort('a', ascending)['a']
